From 616df156cfdf97269e45d6607e52f3d925c46216 Mon Sep 17 00:00:00 2001
From: Melih Mutlu <m.melihmutlu@gmail.com>
Date: Tue, 9 Aug 2022 16:42:23 +0300
Subject: [PATCH] Added pg_buffercache_summary function

Adds pg_buffercache_summary() function into pg_buffercache extension for
retrieving summary information about overall shared_buffer usage.
---
 contrib/pg_buffercache/Makefile               |  3 +-
 .../expected/pg_buffercache.out               |  9 +++
 .../pg_buffercache--1.3--1.4.sql              | 13 +++
 contrib/pg_buffercache/pg_buffercache.control |  2 +-
 contrib/pg_buffercache/pg_buffercache_pages.c | 80 ++++++++++++++++++-
 contrib/pg_buffercache/sql/pg_buffercache.sql |  5 ++
 6 files changed, 108 insertions(+), 4 deletions(-)
 create mode 100644 contrib/pg_buffercache/pg_buffercache--1.3--1.4.sql

diff --git a/contrib/pg_buffercache/Makefile b/contrib/pg_buffercache/Makefile
index d74b3e853c..d6b58d4da9 100644
--- a/contrib/pg_buffercache/Makefile
+++ b/contrib/pg_buffercache/Makefile
@@ -7,7 +7,8 @@ OBJS = \
 
 EXTENSION = pg_buffercache
 DATA = pg_buffercache--1.2.sql pg_buffercache--1.2--1.3.sql \
-	pg_buffercache--1.1--1.2.sql pg_buffercache--1.0--1.1.sql
+	pg_buffercache--1.1--1.2.sql pg_buffercache--1.0--1.1.sql \
+	pg_buffercache--1.3--1.4.sql
 PGFILEDESC = "pg_buffercache - monitoring of shared buffer cache in real-time"
 
 REGRESS = pg_buffercache
diff --git a/contrib/pg_buffercache/expected/pg_buffercache.out b/contrib/pg_buffercache/expected/pg_buffercache.out
index 138556efc9..6994761d0a 100644
--- a/contrib/pg_buffercache/expected/pg_buffercache.out
+++ b/contrib/pg_buffercache/expected/pg_buffercache.out
@@ -8,3 +8,12 @@ from pg_buffercache;
  t
 (1 row)
 
+select used_buffers + unused_buffers > 0,
+        dirty_buffers < used_buffers,
+        pinned_buffers < used_buffers
+from pg_buffercache_summary();
+ ?column? | ?column? | ?column? 
+----------+----------+----------
+ t        | t        | t
+(1 row)
+
diff --git a/contrib/pg_buffercache/pg_buffercache--1.3--1.4.sql b/contrib/pg_buffercache/pg_buffercache--1.3--1.4.sql
new file mode 100644
index 0000000000..02800dbf18
--- /dev/null
+++ b/contrib/pg_buffercache/pg_buffercache--1.3--1.4.sql
@@ -0,0 +1,13 @@
+/* contrib/pg_buffercache/pg_buffercache--1.3--1.4.sql */
+
+-- complain if script is sourced in psql, rather than via ALTER EXTENSION
+\echo Use "ALTER EXTENSION pg_buffercache UPDATE TO '1.4'" to load this file. \quit
+
+CREATE FUNCTION pg_buffercache_summary()
+RETURNS TABLE (used_buffers int4, unused_buffers int4, dirty_buffers int4,
+				pinned_buffers int4, avg_usagecount real)
+AS 'MODULE_PATHNAME', 'pg_buffercache_summary'
+LANGUAGE C PARALLEL SAFE;
+
+-- Don't want these to be available to public.
+REVOKE ALL ON FUNCTION pg_buffercache_summary() FROM PUBLIC;
diff --git a/contrib/pg_buffercache/pg_buffercache.control b/contrib/pg_buffercache/pg_buffercache.control
index 8c060ae9ab..a82ae5f9bb 100644
--- a/contrib/pg_buffercache/pg_buffercache.control
+++ b/contrib/pg_buffercache/pg_buffercache.control
@@ -1,5 +1,5 @@
 # pg_buffercache extension
 comment = 'examine the shared buffer cache'
-default_version = '1.3'
+default_version = '1.4'
 module_pathname = '$libdir/pg_buffercache'
 relocatable = true
diff --git a/contrib/pg_buffercache/pg_buffercache_pages.c b/contrib/pg_buffercache/pg_buffercache_pages.c
index 131bd629b9..6eafa657e6 100644
--- a/contrib/pg_buffercache/pg_buffercache_pages.c
+++ b/contrib/pg_buffercache/pg_buffercache_pages.c
@@ -17,6 +17,7 @@
 
 #define NUM_BUFFERCACHE_PAGES_MIN_ELEM	8
 #define NUM_BUFFERCACHE_PAGES_ELEM	9
+#define NUM_BUFFERCACHE_SUMMARY_ELEM 5
 
 PG_MODULE_MAGIC;
 
@@ -43,7 +44,6 @@ typedef struct
 	int32		pinning_backends;
 } BufferCachePagesRec;
 
-
 /*
  * Function context for data persisting over repeated calls.
  */
@@ -53,12 +53,12 @@ typedef struct
 	BufferCachePagesRec *record;
 } BufferCachePagesContext;
 
-
 /*
  * Function returning data from the shared buffer cache - buffer number,
  * relation node/tablespace/database/blocknum and dirty indicator.
  */
 PG_FUNCTION_INFO_V1(pg_buffercache_pages);
+PG_FUNCTION_INFO_V1(pg_buffercache_summary);
 
 Datum
 pg_buffercache_pages(PG_FUNCTION_ARGS)
@@ -237,3 +237,79 @@ pg_buffercache_pages(PG_FUNCTION_ARGS)
 	else
 		SRF_RETURN_DONE(funcctx);
 }
+
+Datum
+pg_buffercache_summary(PG_FUNCTION_ARGS)
+{
+	Datum		result;
+	TupleDesc	tupledesc;
+	HeapTuple	tuple;
+	Datum		values[NUM_BUFFERCACHE_SUMMARY_ELEM];
+	bool		nulls[NUM_BUFFERCACHE_SUMMARY_ELEM];
+
+	int32 used_buffers = 0;
+	int32 unused_buffers = 0;
+	int32 dirty_buffers = 0;
+	int32 pinned_buffers = 0;
+	float avg_usagecount = 0;
+
+	/* Construct a tuple descriptor for the result rows. */
+	tupledesc = CreateTemplateTupleDesc(NUM_BUFFERCACHE_SUMMARY_ELEM);
+	TupleDescInitEntry(tupledesc, (AttrNumber) 1, "used_buffers",
+						INT4OID, -1, 0);
+	TupleDescInitEntry(tupledesc, (AttrNumber) 2, "unused_buffers",
+						INT4OID, -1, 0);
+	TupleDescInitEntry(tupledesc, (AttrNumber) 3, "dirty_buffers",
+						INT4OID, -1, 0);
+	TupleDescInitEntry(tupledesc, (AttrNumber) 4, "pinned_buffers",
+						INT4OID, -1, 0);
+	TupleDescInitEntry(tupledesc, (AttrNumber) 5, "avg_usagecount",
+						FLOAT4OID, -1, 0);
+
+	BlessTupleDesc(tupledesc);
+
+	for (int i = 0; i < NBuffers; i++)
+	{
+		BufferDesc *bufHdr;
+		uint32		buf_state;
+
+		/*
+		 * No need to get locks on buffer headers as we don't rely on 
+		 * the results in detail. Therefore, we don't get a consistent 
+		 * snapshot across all buffers and it is not guaranteed that 
+		 * the information of each buffer is self-consistent as opposed
+		 * to pg_buffercache_pages.
+		 */
+		bufHdr = GetBufferDescriptor(i);
+		buf_state = pg_atomic_read_u32(&bufHdr->state);
+
+		/* Invalid RelFileNumber means the buffer is unused */
+		if(bufHdr->tag.rlocator.relNumber != InvalidOid)
+		{
+			used_buffers++;
+			avg_usagecount += BUF_STATE_GET_USAGECOUNT(buf_state);
+
+			if (buf_state & BM_DIRTY)
+				dirty_buffers++;
+		}
+		else
+			unused_buffers++;
+
+		if (BUF_STATE_GET_REFCOUNT(buf_state) > 0)
+			pinned_buffers++;
+	}
+	avg_usagecount /= used_buffers;
+
+	memset(nulls, 0, sizeof(nulls));
+	values[0] = Int32GetDatum(used_buffers);
+	values[1] = Int32GetDatum(unused_buffers);
+	values[2] = Int32GetDatum(dirty_buffers);
+	values[3] = Int32GetDatum(pinned_buffers);
+	values[4] = Float4GetDatum(avg_usagecount);
+
+	/* Build and return the tuple. */
+	tuple = heap_form_tuple(tupledesc, values, nulls);
+	result = HeapTupleGetDatum(tuple);
+
+	PG_RETURN_DATUM(result);
+}
diff --git a/contrib/pg_buffercache/sql/pg_buffercache.sql b/contrib/pg_buffercache/sql/pg_buffercache.sql
index e1ba6f7e8d..04e78b1faf 100644
--- a/contrib/pg_buffercache/sql/pg_buffercache.sql
+++ b/contrib/pg_buffercache/sql/pg_buffercache.sql
@@ -4,3 +4,8 @@ select count(*) = (select setting::bigint
                    from pg_settings
                    where name = 'shared_buffers')
 from pg_buffercache;
+
+select used_buffers + unused_buffers > 0,
+        dirty_buffers < used_buffers,
+        pinned_buffers < used_buffers
+from pg_buffercache_summary();
\ No newline at end of file
-- 
2.25.1

